/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/
/*
 * file.c
 *
 * This file contains the routines that load and discard voms-proxy-init
 * configuration file.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

#include <string>
#include <vector>
#include <iostream>
#include <fstream>

#include "file.h"

/*
 * This takes into account the fact that there won't (hopefully) be thousands
 * of voms servers registered...
 */
static std::vector<vomsdata> array;

static inline bool tokenize(std::string str, unsigned int &start, std::string &value);
static bool loadfile0(std::string filename, uid_t uid, gid_t gid, bool debug, bool wasempty);


/*
 * Function:
 *   discardfile()
 *
 * Description:
 *   This function discards the file loaded in memory by loadfile()
 *
 * Parameters:
 *   NONE.
 *
 * Result:
 *   NONE.
 */
void 
discardfile(void)
{
  array.clear();
}

/*
 * Function:
 *   tokenize(str)
 *
 * Description:
 *   This function separates a std::string into "-delimited fields, and
 *   returns them one at a time.
 *
 * Parameters:
 *   'str' - The std::string from which to extract the field. This may
 *           be NULL, meaning that you want the next field from the
 *           same std::string used in the previous calls. The std::string itself
 *           is modified by this call.
 *
 * Result:
 *   Failure:
 *     NULL.
 *   Success:
 *     The std::string that describes the parameter.
 */
static inline bool
tokenize(std::string str, unsigned int &start, std::string &value)
{
  if (start != std::string::npos) {
    unsigned int begin = str.find('"',start);
    if (begin != std::string::npos) {
      unsigned int end = str.find('"',begin+1);
      if (end != std::string::npos) {
	value = str.substr(begin+1, end-begin-1);
	start = end+1;
	if (start >= str.size())
	  start = std::string::npos;
	return true;
      }
    }
  }
  return false;
}

/*
 * Function:
 *   loadfile(filename)
 *
 * Description:
 *   This function loads a configration file into memory.
 *
 * Parameters:
 *   'filename' - The name of the file to load. There are some specific
 *                requirements about this file: It must be a regular file,
 *                its permissions must be 0644, and it must belong to user 0
 *                and group 0 (root.root).
 *
 * Result:
 *   Failure:
 *     0.
 *   Success:
 *     1.
 */
bool loadfile(std::string filename, uid_t uid, gid_t gid, bool debug, bool wasempty)
{
  struct stat stats;

  int error = 0;

  struct vomsdata data;

  if (filename.empty()) {
    if (!wasempty || debug)
      std::cerr << "Configuration filename unspecified!" << std::endl << std::flush;
    return false;
  }


  if (stat(filename.c_str(), &stats) == -1 || stats.st_uid != uid ||
      stats.st_gid != gid ||
      !(stats.st_mode == (S_IFREG | S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) ||
       stats.st_mode == (S_IFDIR |
			 S_IRUSR | S_IWUSR | S_IXUSR |
			 S_IRGRP | S_IXGRP |
			 S_IROTH | S_IXOTH))) {
    if (!wasempty || debug)
      std::cerr << "Invalid configuration filename: " << filename << std::endl <<std::flush;
    return false;
  }

  if (stats.st_mode & S_IFREG)
    return loadfile0(filename, uid, gid, debug, wasempty);
  else {
    if (debug)
      std::cout << "Open configuration dir: " << filename << std::endl << std::flush;
    DIR *dp = opendir(filename.c_str());
    struct dirent *de;
    bool cumulative = false;

    if (dp) {
      while (de = readdir(dp)) {
	char *name = de->d_name;
	if (name && (strcmp(name, ".") != 0) && (strcmp(name, "..") != 0))
	  cumulative |= loadfile(filename + "/" + name, uid, gid, debug, wasempty);
      }
      return cumulative;
    }
    else {
      if (!wasempty || debug)
	std::cerr << "Unable to open dir " << filename << std::endl << std::flush;
      return false;
    }
  }
}

static bool loadfile0(std::string filename, uid_t uid, gid_t gid, bool debug, bool wasempty)
{
  int error = 0;

  struct vomsdata data;

  if (filename.empty()) {
    if (!wasempty || debug)
      std::cerr << "Configuration filename unspecified!" << std::endl << std::flush;
    return false;
  }

  /* Opens the file */
  std::ifstream f(filename.c_str());

  if (!f) {
    if (!wasempty || debug)
      std::cerr << "Unable to open configuration file: " << filename << "!" << std::endl << std::flush;
    return false;
  }
  else if (debug)
    std::cout << "Open configuration file: " << filename << std::endl << std::flush;

  /* Load the file */
  int linenum = 1;
  bool ok = true;

  while (ok && f) {
    std::string line;

    if (getline(f,line)) {
      unsigned int start = 0;
      std::string port;

      ok &= tokenize(line, start, data.nick);
      ok &= tokenize(line, start, data.host);
      ok &= tokenize(line, start, port);
      ok &= tokenize(line, start, data.contact);
      ok &= tokenize(line, start, data.vo);

      if (!ok)
	std::cerr << "Error on reading or parsing line " << linenum 
	     << " file: " << filename << std::endl << std::flush;
      else {
	data.port = atoi(port.c_str());
	array.push_back(data);
      }

      if (debug)
	std::cout << data.nick << " (" << data.host << ":" << port << " ("
		  << data.vo << ")) " << data.contact << std::endl << std::flush;

    }
    linenum++;
  }

  /*
   * No error and file completely read.
   */

  if (!error && f.eof())
      return true;

  discardfile();
  return false;
}

/*
 * Function:
 *   findinfile(nick, host, port, contact, vo)
 *
 * Description:
 *   This function searches the file loaded by loadfile() for a specific
 *   nick, and initializes the other three parameters consequently. The
 *   search is sequential, but because the expected number of VOMS servers
 *   is not very high (at most a score or so), this is no problem.
 *
 * Parameters:
 *   'nick'    - The nick of the VOMS server to find.
 *   'host'    - OUTPUT PARAMETER: The hostname of the VOMS server.
 *   'port'    - OUTPUT PARAMETER: The port on which the VOMS server listens.
 *   'contact' - OUTPUT PARAMETER: The subject of the VOMS server's
 *               certificate.
 *   'vo'      - OUTPUT PARAMETER: The name of the VO to whom he VOMS server
 *               belongs.
 *
 * Result:
 *   Failure:
 *     0.
 *   Success:
 *     Non zero. Furthermore, the four output parameters are initialized
 *     properly.
 */
std::vector<vomsdata>
findinfile(std::string nick)
{
  std::vector<vomsdata>::iterator beg = array.begin(), end = array.end();
  std::vector<vomsdata> results;

  while (beg != end) {
    if (beg->nick ==  nick)
      results.push_back(*beg);
    beg++;
  }

  return results;
}
